#ifndef CDictionaryIntToInt_H
#define CDictionaryIntToInt_H

#include <iostream>
#include <map>

#ifdef SET
#include <set>
#elif defined QUEUE
#include <queue>
#include <list>
#elif defined UNORDERED_MULTIMAP
#include <unordered_map>
#elif not defined MAP
#include <forward_list>
#include <vector>
#include <algorithm>
#endif

template<typename KeyType, typename ValueType>
class CDictionaryIntToInt {
#ifdef MAP
    std::map<KeyType, ValueType> data;
#elif defined SET
    std::set<std::pair<KeyType, ValueType>> data;
#elif defined UNORDERED_MULTIMAP
    std::unordered_multimap<KeyType, ValueType> data;
#elif defined QUEUE
    std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> data;
#else
    std::forward_list<std::pair<KeyType, ValueType>> data;
#endif
    int total = 0;
    
public:
    CDictionaryIntToInt() = default;
    CDictionaryIntToInt(KeyType key, ValueType value) {
        add(key, value);
    }
    
    bool add(KeyType key, ValueType value, char force = 0) {
#ifdef MAP
        if (!force && data.find(key) != data.end()) {
            throw std::runtime_error("Ключ вже зареєстровано");
        }
        data[key] = value;
        total++;
        return true;
#elif defined SET
        auto it = std::find_if(data.begin(), data.end(), [&](const std::pair<KeyType, ValueType>& elem) {
            return elem.first == key;
        });
        if (it != data.end()) {
            throw std::runtime_error("Ключ вже зареєстровано");
        }
        data.emplace(key, value);
        total++;
        return true;
#elif defined UNORDERED_MULTIMAP
        auto it = data.find(key);
        if (it != data.end()) {
            throw std::runtime_error("Ключ вже зареєстровано");
        }
        data.emplace(key, value);
        total++;
        return true;
#elif defined QUEUE
        if (!force) {
            std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> temp = data;
            while (!temp.empty()) {
                auto elem = temp.front();
                temp.pop();
                if (elem.first == key) {
                    throw std::runtime_error("Ключ вже зареєстровано");
                }
            }
        }
        total++;
        data.push(std::make_pair(key, value));
        return true;
#else
        auto it = std::find_if(data.begin(), data.end(), [&](const std::pair<KeyType, ValueType>& elem){
            return elem.first == key;
        });
        if (it != data.end() && !force)
            throw std::runtime_error("Ключ вже зареєстровано");
        if (it == data.end() || force) {
            data.push_front(std::make_pair(key, value));
            ++total;
            return true;
        }
        return false;
#endif
    }
    
    ValueType get(KeyType key) const {
#ifdef MAP
        auto it = data.find(key);
        if (it != data.end()) {
            return it->second;
        }
#elif defined SET
        auto it = std::find_if(data.begin(), data.end(), [&](const std::pair<KeyType, ValueType>& elem) {
            return elem.first == key;
        });
        if (it != data.end()) {
            return it->second;
        }
#elif defined UNORDERED_MULTIMAP
        auto it = data.find(key);
        if (it != data.end()) {
            return it->second;
        }
#elif defined QUEUE
        std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> temp = data;
        while (!temp.empty()) {
            auto elem = temp.front();
            temp.pop();
            if (elem.first == key) {
                return elem.second;
            }
        }
#else
        std::vector<std::pair<KeyType, ValueType>> sorted_data(data.begin(), data.end());
        std::sort(sorted_data.begin(), sorted_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
            return a.first < b.first;
        });
        auto it = std::lower_bound(sorted_data.begin(), sorted_data.end(), key, [](const std::pair<KeyType, ValueType>& elem, KeyType key) {
            return elem.first < key;
        });
        if (it != sorted_data.end() && it->first == key) {
            return it->second;
        }
#endif
        throw std::out_of_range("Key not found");
    }
    
    bool remove(KeyType key) {
#if defined(MAP) || defined(UNORDERED_MULTIMAP)
        if(data.erase(key) > 0){total--;
            return 1;}
#elif defined SET
        auto it = std::find_if(data.begin(), data.end(), [&](const std::pair<KeyType, ValueType>& elem) {
            return elem.first == key;
        });
        if (it != data.end()) {
            data.erase(it);
            total--;
            return true;
        }
        return false;
#elif defined QUEUE
        std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> new_data;
        bool removed = false;
        while (!data.empty()) {
            auto elem = data.front();
            data.pop();
            if (elem.first != key || removed) {
                new_data.push(elem);
            } else removed = true;
        }
        data = std::move(new_data);
        if(removed){total--;return 1;}else return 0;
#else
        auto prev = data.before_begin();
        for (auto it = data.begin(); it != data.end(); ++it) {
            if (it->first == key) {
                data.erase_after(prev);
                --total;
                return true;
            }
            prev = it;
        }
        return false;
#endif
    }
    
    CDictionaryIntToInt operator+(const CDictionaryIntToInt& other) const {
        CDictionaryIntToInt result(*this);
#ifdef MAP
        for (const auto& [key, value] : other.data) {
            result.data.emplace(key, value);
        }
#elif defined SET
        std::set<std::pair<KeyType, ValueType>> union_set;
        std::set_union(data.begin(), data.end(), other.data.begin(), other.data.end(), std::inserter(union_set, union_set.begin()));
        result.data = union_set;
#elif defined UNORDERED_MULTIMAP
        for (const auto& elem : other.data) {
            result.data.emplace(elem.first, elem.second);
        }
#elif defined QUEUE
        std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> temp = other.data;
        while (!temp.empty()) {
            auto elem = temp.front();
            temp.pop();
            result.add(elem.first, elem.second);
        }
#else
        std::vector<std::pair<KeyType, ValueType>> all_data(data.begin(), data.end());
        all_data.insert(all_data.end(), other.data.begin(), other.data.end());
        std::sort(all_data.begin(), all_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
            return a.first < b.first;
        });
        all_data.erase(std::unique(all_data.begin(), all_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
                           return a.first == b.first;
                       }), all_data.end());
        result.data.assign(all_data.begin(), all_data.end());
        result.total = all_data.size();
#endif
        return result;
    }
    
    CDictionaryIntToInt operator-(const CDictionaryIntToInt& other) const {
        CDictionaryIntToInt result;
#ifdef MAP
        for (const auto& [key, value] : data) {
            if (other.data.find(key) == other.data.end()) {
                result.data.emplace(key, value);
            }
        }
#elif defined SET
        std::set<std::pair<KeyType, ValueType>> diff_set;
        std::set_difference(data.begin(), data.end(), other.data.begin(), other.data.end(), std::inserter(diff_set, diff_set.begin()));
        result.data = diff_set;
#elif defined UNORDERED_MULTIMAP
        for (const auto& elem : data) {
            if (other.data.find(elem.first) == other.data.end()) {
                result.data.emplace(elem.first, elem.second);
            }
        }
#elif defined QUEUE
        std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> temp = data;
        while (!temp.empty()) {
            auto elem = temp.front();
            temp.pop();
            if (other.get(elem.first) != elem.second) {
                result.add(elem.first, elem.second);
            }
        }
#else
        std::vector<std::pair<KeyType, ValueType>> sorted_data(data.begin(), data.end());
        std::vector<std::pair<KeyType, ValueType>> other_sorted_data(other.data.begin(), other.data.end());
        std::sort(sorted_data.begin(), sorted_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
            return a.first < b.first;
        });
        std::sort(other_sorted_data.begin(), other_sorted_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
            return a.first < b.first;
        });
        std::vector<std::pair<KeyType, ValueType>> diff_set;
        std::set_difference(sorted_data.begin(), sorted_data.end(), other_sorted_data.begin(), other_sorted_data.end(), std::back_inserter(diff_set));
        result.data.assign(diff_set.begin(), diff_set.end());
        result.total = diff_set.size();
#endif
        return result;
    }
    
    CDictionaryIntToInt operator*(const CDictionaryIntToInt& other) const {
        CDictionaryIntToInt result;
#ifdef MAP
        for (const auto& [key, value] : data) {
            auto it = other.data.find(key);
            if (it != other.data.end()) {
                result.data.emplace(key, value);
            }
        }
#elif defined SET
        std::set<std::pair<KeyType, ValueType>> intersection_set;
        std::set_intersection(data.begin(), data.end(), other.data.begin(), other.data.end(), std::inserter(intersection_set, intersection_set.begin()));
        result.data = intersection_set;
#elif defined UNORDERED_MULTIMAP
        for (const auto& elem : data) {
            auto range = other.data.equal_range(elem.first);
            for (auto it = range.first; it != range.second; ++it) {
                if (elem.second == it->second) {
                    result.data.emplace(elem.first, elem.second);
                }
            }
        }
#elif defined QUEUE
        std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> temp = data;
        while (!temp.empty()) {
            auto elem = temp.front();
            temp.pop();
            if (other.get(elem.first) == elem.second) {
                result.add(elem.first, elem.second);
            }
        }
#else
        std::vector<std::pair<KeyType, ValueType>> sorted_data(data.begin(), data.end());
        std::vector<std::pair<KeyType, ValueType>> other_sorted_data(other.data.begin(), other.data.end());
        std::sort(sorted_data.begin(), sorted_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
            return a.first < b.first;
        });
        std::sort(other_sorted_data.begin(), other_sorted_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
            return a.first < b.first;
        });
        std::vector<std::pair<KeyType, ValueType>> intersection_set;
        std::set_intersection(sorted_data.begin(), sorted_data.end(), other_sorted_data.begin(), other_sorted_data.end(), std::back_inserter(intersection_set));
        result.data.assign(intersection_set.begin(), intersection_set.end());
        result.total = intersection_set.size();
#endif
        return result;
    }
    
    CDictionaryIntToInt& operator=(const CDictionaryIntToInt& other) = default;
    
    bool operator==(const CDictionaryIntToInt& other) const {
#ifdef MAP
        return data == other.data;
#elif defined SET
        return data == other.data;
#elif defined UNORDERED_MULTIMAP
        return data == other.data;
#elif defined QUEUE
        std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> temp1 = data;
        std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> temp2 = other.data;
        while (!temp1.empty() && !temp2.empty()) {
            if (temp1.front() != temp2.front()) {
                return false;
            }
            temp1.pop();
            temp2.pop();
        }
        return temp1.empty() && temp2.empty();
#else
        if (total != other.total) return false;
        std::vector<std::pair<KeyType, ValueType>> sorted_data(data.begin(), data.end());
        std::vector<std::pair<KeyType, ValueType>> other_sorted_data(other.data.begin(), other.data.end());
        std::sort(sorted_data.begin(), sorted_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
            return a.first < b.first;
        });
        std::sort(other_sorted_data.begin(), other_sorted_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
            return a.first < b.first;
        });
        return sorted_data == other_sorted_data;
#endif
    }
    
    int getTotal() const {
#ifdef MAP
        return data.size();
#elif defined SET
        return data.size();
#elif defined UNORDERED_MULTIMAP
        return data.size();
#elif defined QUEUE
        return data.size();
#else
        return total;
#endif
    }
    
    template<typename K, typename V>
    friend std::istream& operator>>(std::istream& in, CDictionaryIntToInt<K, V>& dict);
    
    template<typename K, typename V>
    friend std::ostream& operator<<(std::ostream& out, const CDictionaryIntToInt<K, V>& dict);
    std::map<KeyType, ValueType> getMapWithKey(KeyType key) const {
    std::map<KeyType, ValueType> result;
        ValueType value = get(key);
        result[key] = value;
    return result;
};
};

template<typename KeyType, typename ValueType>
std::istream& operator>>(std::istream& in, CDictionaryIntToInt<KeyType, ValueType>& dict) {
    KeyType key;
    ValueType value;
    while (in >> key >> value) {
        if (!dict.add(key, value)) {
            throw std::runtime_error("Ключ \"" + std::to_string(key) + "\" вже було введено");
        }
    }
    return in;
}

template<typename KeyType, typename ValueType>
std::ostream& operator<<(std::ostream& out, const CDictionaryIntToInt<KeyType, ValueType>& dict) {
#ifdef MAP
    for (const auto& [key, value] : dict.data) {
        out << key << " -> " << value << std::endl;
    }
#elif defined SET
    for (const auto& elem : dict.data) {
        out << elem.first << " -> " << elem.second << std::endl;
    }
#elif defined UNORDERED_MULTIMAP
    for (const auto& [key, value] : dict.data) {
        out << key << " -> " << value << std::endl;
    }
#elif defined QUEUE
    std::queue<std::pair<KeyType, ValueType>, std::list<std::pair<KeyType, ValueType>>> temp = dict.data;
    while (!temp.empty()) {
        auto elem = temp.front();
        temp.pop();
        out << elem.first << " -> " << elem.second << std::endl;
    }
#else
    std::vector<std::pair<KeyType, ValueType>> sorted_data(dict.data.begin(), dict.data.end());
    std::sort(sorted_data.begin(), sorted_data.end(), [](const std::pair<KeyType, ValueType>& a, const std::pair<KeyType, ValueType>& b) {
        return a.first < b.first;
    });
    for (const auto& elem : sorted_data) {
        out << elem.first << " -> " << elem.second << std::endl;
    }
#endif
    return out;
}

#endif // CDictionaryIntToInt_H
